#
#   COMPONENT_NAME: LINUXDRV
#
#   FUNCTIONS: none
#
#   ORIGINS:
#
#
#   (C) COPYRIGHT International Business Machines Corp. 1996
#   All Rights Reserved
#   Licensed Materials - Property of IBM
#
#
#  PERL package
#
#  Package      : Support
# 
#  Description	: This package provide a hierarchy or classes which handle
#                 PAF files  .
#                 This package  is  designed to be used by 'IBMnpmenu' and 
#                 'IBMprinttool' tools .
#

#################################################################################
#################################################################################
#
#                           package  Support::base
#
#################################################################################
#################################################################################
package Support::base;

use strict;

#________________________________________________________________________________
#
#  package:     Support::base
#
#  subroutine:  new
#   
#  description: the constructor of the class   
#
#  parameters: PAFfilename  - the full name  of the PAF file 
#
#  return:      the reference to the created object 
#
#________________________________________________________________________________
#	
sub new {

    my $class = shift;
    my $this  = {};
    
    if(  @_ ) {
	$this->{ PAFfilename } = shift;
    }
    else {
	die " Error !!! $class : PAFfilename required ";
    }

    bless( $this , $class );
    return $this;

}

#________________________________________________________________________________
#
#  package:     Support::base
#
#  subroutine:  isValid
#   
#  description: check if the PAF file is valid    
#
#  parameters:  ( none )
#
#  return:      1   - if the PAF file is valid
#               0   - if the PAF file is not valid 
#
#________________________________________________________________________________
#
sub isValid {
    my $this = shift;
    my $PAFfilename = $this->{PAFfilename};
    my @PAFstat;
    my $PAFtime;    
    # check if the file exists
    if (! -e $PAFfilename ) {
	return 0;
    }

    # ########################
    # for debuging !!!!!!
    return 1;
    # ########################


    # get the date:time of the last modification
    # of PAF file
    @PAFstat=stat($PAFfilename);
    $PAFtime=$PAFstat[9];

    # check if the PAF file is not corrupted

    open ( IN_PAF , $PAFfilename );
    my $crc_line=<IN_PAF>;
    close( IN_PAF );
    chop($crc_line);
    my $crc=substr($crc_line,3);

    if ( $PAFtime > $crc+100 ) {
	warn "...$PAFfilename ... \n";
	return 0;
    }

    # check if the date:time of PAF  is earlier than PPD file

}


#________________________________________________________________________________
#
#  package:     Support::base
#       
#  subroutine:  eject_escape
#   
#  description: replace perl-escape sequences from the string
#
#  parameters:  the string
#
#  return:      the string without escape sequences
#
#________________________________________________________________________________
#
sub eject_escape {

    my $class = shift;
    my $string = shift;

    my $found ="";
          
    $_ = $string;
    if( m/&#[\d]{3};/ ) {
    $found = $&;
    }
    else {
  $found = "";
    }
    while ( length($found) ) {
  $_ = $found;
  m/[\d]{3}/; 
  
  my $value = chr ($&);
        $string =~  s/$found/$value/;
  $_ = $string;
        if ( m/&#[\d]{3};/ ) {
       $found = $&;
   }
  else {
      $found = "";
  }
    }
    $string;
    
} 



#################################################################################
#################################################################################
#
#                  package  Support::messages
#
#################################################################################
#################################################################################
package Support::messages;

use strict;
use vars qw( @ISA );

@ISA = qw( Support::base );


#________________________________________________________________________________
#
#  package:     Support::messages
#
#  subroutine:  new
#   
#  description: the constructor of the class   
#
#  parameters:  
#               PAFfilename  - the full name  of the PAF file 
#
#  return:      the reference to the created object 
#
#________________________________________________________________________________
#	
sub new {
    my $class = shift;

    my $this=new Support::base( @_ );

    my $trans = {};

    $this->{trans} = $trans;

    bless($this,$class);

    $this->fillFromPAF;

    return $this;   
}

#________________________________________________________________________________
#
#  package:     Support::messages
#
#  subroutine:  fillFromPAF
#   
#  description: fill data members of the class form PAF 
#
#  parameters:  ( no parameters )
#
#  return:      ( no value )
#________________________________________________________________________________
#
sub fillFromPAF {
    my $this = shift;
    my $PAFfilename = $this->{PAFfilename};
    my $trans = {}; 
    $this->{trans} = $trans;

    my $key = undef;
    my $value;
    my $language = undef;

    my @buffer;
    open ( IN_PAF , $PAFfilename );
    @buffer=<IN_PAF>;
    close( IN_PAF );

    # start to parse the buffer ...
    chop(@buffer);
    $key = undef;
    my $status = "undef";         #  this can be "undef" , "lang" or "trans" 
    foreach    my $item  (@buffer) {
        # out with blank lines 
	if ( $item eq '' ) {
	    next;
	}
        # out with comments 
	if ( substr($item,0,2) eq '##' ) {
	    next;
	}
        # start to parse translations
	if ( substr($item,0,9) eq '#language' ) {      # start a new language
            # get ready for start a new language
	    $status = "lang";
	    $key = "language";
	    next;
	}
	if ( substr($item,0,13) eq '#end_language' ) { # end the curent language
            # get ready to save the language
	    $status = "undef";
	    next;
	}

	if ( substr($item,0,13) eq '#translations' ) { 
            # get ready to fill in the translation hash
	    $status = "trans";
	    next;
	}

	if ( substr($item,0,17) eq '#end_translations' ) { 
	    $status = "lang";
	    next;
	}


       if (!  $key  ) {
           $key = substr($item,2);
           next;
        }
	else {
	    $value=$item;
	    if( $status eq "lang" ) {
		$this->{$key}=$value;
	    }
	    else {
		$trans->{$key}=$value;
	    }
	    $key = undef;
	}
 
    } 
    
}



#________________________________________________________________________________
#
#  package:     Support::messages
#
#  subroutine:  translate
#   
#  description: translate a text in the object language 
#
#  parameters:  text - the text that will be translated
#
#  return:      ( no value )
#________________________________________________________________________________
#
sub translate {
    my $this = shift;
    my $PAFfilename = $this->{PAFfilename};
    my $trans = $this->{trans};
    my $result = undef;

    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter .\n");
    }
    my $key = shift; 
    $result=$trans->{$key};

    if (! $result ){
	warn("Support::messages : no translation found for \'$key\'",
	     "return :  $key \n");
	$result = $key;
    }

    return $result;
}



#################################################################################
#################################################################################
#
#                  package  Support::ibmXXXXX
#
#################################################################################
#################################################################################
package Support::ibmXXXXX;

use strict;
use vars qw( @ISA );

@ISA = qw( Support::base );

#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  new
#   
#  description: the constructor of the class   
#
#  parameters: PAFfilename  - the full name  of the PAF file 
#
#  return:      the reference to the created object 
#
#________________________________________________________________________________
#	
sub new {
    my $class=shift;
    my $this=new Support::base(@_);
    
    # instalable options UIs list
    my @installuis=();
    # paper input/output UIs list
    my @uis1=();
    # other  UIs list
    my @uis2=();
    # jcls  UIs list
    my @jcluis=();
    # constraints list
    my @constraints=();
    # custom paper size hash
    my $customPaperSize=undef;

    $this->{installuis}=\@installuis;
    $this->{uis1}=\@uis1;
    $this->{uis2}=\@uis2;
    $this->{jcluis}=\@jcluis;

    $this->{constraints}=\@constraints;
    $this->{customPaperSize}=$customPaperSize;
    bless($this,$class);

    if ( ! $this->isValid ) {
	# it should be recreated somehow form ppd file ...
    }
    else {
	$this->fillFromPAF;
    }
    return $this;   
}


#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  fillFromPAF
#   
#  description: fill data members of the class form PAF 
#
#  parameters:  ( no parameters )
#
#  return:      ( no value )
#________________________________________________________________________________
#
sub fillFromPAF {

    my $this = shift;
    my $PAFfilename = $this->{PAFfilename};

    my $installuis      =$this->{installuis};
    my $uis1            =$this->{uis1};
    my $uis2            =$this->{uis2};
    my $jcluis          =$this->{jcluis};
    my $constraints     =$this->{constraints};
    my $customPaperSize =$this->{customPaperSize}; 

    @$installuis  =();
    @$uis1        =();
    @$uis2        =();
    @$jcluis      =();
    @$constraints =();

    my $currentFeature=undef;
    my $currentOptions=undef;
    my $currentOption=undef;

    my $key = undef;
    my $value;

    my @buffer;
    my $status = "undef" ; # it can be "param" , "options" , "option" , "undef"
    open ( IN_PAF , $PAFfilename );
    @buffer=<IN_PAF>;
    close( IN_PAF );

    # start to parse the buffer ...
    chop(@buffer);
    
    $status = "undef";
    foreach    my $item  (@buffer) {
        # out with blank lines 
	if ( $item eq '' || substr($item,0,1) eq ' ') {
	    next;
	}
        # out with comments 
	if ( substr($item,0,2) eq '##' ) {
	    next;
	}

	if (substr($item,0,4) eq "#ccc") {
	    #
	    # CONSTRAINT
	    #
	    $status = "param";
	    #get ready to start a new constraint
	    $currentFeature = {};
	    @$constraints=(@$constraints,$currentFeature);
	    next;
        } 

	# end the constraint specification
	if (substr($item,0,8) eq "#end_ccc") {
	    #
	    # CONSTRAINT
	    #
	    $status = "undef";
	    $currentFeature = undef;
	    next;
        } 


	# detect what kind of information is !!!
	if ( substr($item,0,5) eq "#uiio") {
	    #
	    # INSTALLABLE OPTIONS 
	    #
	    $status = "param";
            # get ready for start a new ui
	    $currentFeature = {};
	    @$installuis=(@$installuis,$currentFeature);
	    next;
	}
	if (substr($item,0,4) eq "#uip") {
	    #
	    # PAPER RELATED UIs
	    #
	    $status = "param";
	    # get ready to start a new ui
	    $currentFeature = {};
	    @$uis1=(@$uis1,$currentFeature);
	    next;
	}
	if (substr($item,0,4) eq "#uio") {
	    #
	    # OTHER UIs
	    #
	    $status = "param";
	    # get ready to start a new ui
	    $currentFeature = {};
	    @$uis2=(@$uis2,$currentFeature);
	    next;
	}
	if (substr($item,0,4) eq "#uij") {
	    #
	    # JCL UIs
	    #
	    $status = "param";
	    # get ready to start a new ui
	    $currentFeature = {};
	    @$jcluis=(@$jcluis,$currentFeature);
	    next;
	}
	# the end of the ui specification
	if ( substr($item,0,7) eq '#end_ui' ) {
	    $status = "undef";
	    $currentFeature=undef;
 	    next;
	}
  
	#### parse options for UIs
	if ( substr($item,0,9) eq '#_options' ) {
	    $status = "_options";
	    # get ready for starting options
	    my @list_options=();
	    $currentOptions=\@list_options;
	    $currentFeature->{options}=$currentOptions;
	    next;
	}
	# end the options section
	if ( substr($item,0,12) eq '#end_options' ) {
	    $status = "param";
	    next;
	}

	# detect the begining of an option
	if (substr($item,0,8) eq '#_option') {
	    $status = "option";
	    # get ready to get the option
	    $currentOption={};
	    @$currentOptions = ( @$currentOptions , $currentOption );	    
	    next;
	}
	# detect the end of an option
	if (substr($item,0,11) eq '#end_option') {
	    $status = "options";
	    next;
	}

	if (substr($item,0,4) eq "#cps") {
	    #
	    # CUSTOM PAPER SIZE
	    #
	    $status = "cps";
	    #get ready to start the custom paper size hash
	    $customPaperSize = {};
	    next;
        } 

	# end the custom paper size  specification
	if (substr($item,0,8) eq "#end_cps") {
	    #
	    #  CUSTOM PAPER SIZE
	    #
	    $this->{customPaperSize}=$customPaperSize;    
	    $status = "undef";
	    next;
        } 

	
	# set  ( key , value ) pairs !!!!
	if (!  $key  ) {
	    $key = substr($item,1);
	    next;
        }

        $value=$item;
        if($status eq "param") {
	    $currentFeature->{$key}=$value;
	    $key = undef;
        }
	elsif( $status eq "cps" ) {
	    $customPaperSize->{$key}=$value;
	    $key = undef;
	} else {
	    $currentOption->{$key}=$value;
	    $key = undef;
	}
    }  

    # put the translations for the driver features/options
    my $ibmDataDir = '/etc/ibmlp/data';
    # PAF folder
    my $ibmPafDir = "${ibmDataDir}/pafs";
    # files which contain messages  ( PAF format )
    my $ibmMessagesPaf = "${ibmPafDir}/messages.paf";
    # create the translation object
    my $transObj = new Support::messages($ibmMessagesPaf);
    $this->setFeatureTranslation("*CollateLocation",
				 $transObj->translate("CollateFeature"));
    $this->setFeatureOptionTranslation("*CollateLocation",
				       "Printer",
				 $transObj->translate("CollateOptionPrinter"));
    $this->setFeatureOptionTranslation("*CollateLocation",
				       "Driver",
				 $transObj->translate("CollateOptionDriver"));


}


#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  enumFeatures
#   
#  description: returns an array with Features IDs
#               all UIs :
#                      - installable options UIs
#                      - input/paper/outpui UIs
#                      - other UIs
#               WARNING:
#                      - no other information from PAF !!!!!!
#
#  parameters:  ( no parametters )
#
#  return:     an array with all Featues IDs   if there exists any feature
#               otherways  a blank array .
#________________________________________________________________________________
#
sub enumFeatures {
    my $this = shift;

    my $installuis   =$this->{installuis};
    my $uis1         =$this->{uis1};
    my $uis2         =$this->{uis2};



    my @result = ();

    foreach my $anonimus (@$installuis) {
	@result = ( @result , $anonimus->{id} );
    }
    
    foreach my $anonimus (@$uis1) {
	@result = ( @result , $anonimus->{id} );
    }
    
    foreach my $anonimus (@$uis2) {
	@result = ( @result , $anonimus->{id} );
    }

    return @result;
}


#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  enumJclFeatures
#   
#  description: returns an array with Jcl Features IDs
#
#               WARNING:
#                      - no other information from PAF !!!!!!
#
#  parameters:  ( no parametters )
#
#  return:     an array with all Jcl Featues IDs   if there exists any feature
#               otherways  a blank array .
#________________________________________________________________________________
#
sub enumJclFeatures {
    my $this = shift;

    my $jcluis         =$this->{jcluis};


    my @result = ();

    foreach my $anonimus (@$jcluis) {
	@result = ( @result , $anonimus->{id} );
    }
    
    return @result;
}


#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  enumInstallableOptions
#   
#  description: returns an array with installable options UIs
#
#  parameters:  ( no parametters )
#
#  return:     an array with all installable options IDs   
#              if there exists any feature otherways  a blank array .
#________________________________________________________________________________
#
sub enumInstallableOptions {
    my $this = shift;

    my $installuis   =$this->{installuis};

    my @result = ();

    foreach my $anonimus (@$installuis) {
	@result = ( @result , $anonimus->{id} );
    }
    
    return @result;
}


#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  enumIOPaperUIs
#   
#  description: returns an array with input/output/paper UIs
#
#  parameters:  ( no parametters )
#
#  return:     an array with all input/output/paper IDs   
#              if there exists any feature otherways  a blank array .
#________________________________________________________________________________
#
sub enumIOPaperUIs {
    my $this = shift;

    my $uis1   =$this->{uis1};

    my @result = ();

    foreach my $anonimus (@$uis1) {
	@result = ( @result , $anonimus->{id} );
    }
    
    return @result;
}


#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  enumOtherUIs
#   
#  description: returns an array with all UIs which are not in :
#                     -  input/output/paper UIs category
#                     -  installable options UIs category
#
#  parameters:  ( no parametters )
#
#  return:     an array with all other  IDs   
#              if there exists any feature otherways  a blank array .
#________________________________________________________________________________
#
sub enumOtherUIs {
    my $this = shift;

    my $uis2   =$this->{uis2};

    my @result = ();

    foreach my $anonimus (@$uis2) {
	@result = ( @result , $anonimus->{id} );
    }
    
    return @result;
}

#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  enumAllUIs
#   
#  description: returns an array with all UIs which are not in :
#                     -  installable options UIs category
#
#  parameters:  ( no parametters )
#
#  return:     an array with all other  IDs   
#              if there exists any feature otherways  a blank array .
#________________________________________________________________________________
#
sub enumAllUIs {
    my $this = shift;

    my $uis2   =$this->{uis2};
    my $uis1   =$this->{uis1};

    my @result = ();

    foreach my $anonimus (@$uis1) {
	@result = ( @result , $anonimus->{id} );
    }

    foreach my $anonimus (@$uis2) {
	@result = ( @result , $anonimus->{id} );
    }
    
    return @result;
}



#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  setFeatureTranslation
#   
#  description: sets the translation of a Feature of cource if that 
#               feature exists
#               is used to set translations for features which
#               re defined in the driver and are not available from PPD file
#
#  parameters:  featureID - the ID of the Feature
#               translation - the translation 
#
#  return:      none
#________________________________________________________________________________
#
sub setFeatureTranslation {
    my $this = shift;
    my $result = undef;

    my $installuis   =$this->{installuis};
    my $uis1         =$this->{uis1};
    my $uis2         =$this->{uis2};
    my $jcluis       =$this->{jcluis};

    # check if the function has two parameter passed
    if ( @_ < 2 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 2 parameter .\n");
    }

    my $featureID = shift;
    my $translation = shift;
    
    foreach my $anonimus (@$installuis) {
	if( $anonimus->{id} eq $featureID ) {
	    $anonimus->{translation}=$translation;
	    return ;
	}
    }
    foreach my $anonimus (@$uis1) {
	if( $anonimus->{id} eq $featureID ) {
	    $anonimus->{translation}=$translation;
	    return ;
	}
    }
    foreach my $anonimus (@$uis2) {
	if( $anonimus->{id} eq $featureID ) {
	    $anonimus->{translation}=$translation;
	    return ;
	}
    }
    foreach my $anonimus (@$jcluis) {
	if( $anonimus->{id} eq $featureID ) {
	    $anonimus->{translation}=$translation;
	    return ;
	}
    }

}


#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  getFeatureTranslation
#   
#  description: returns a value that contains translation of a  feature 
#               of course if there exists any translation 
#
#  parameters:  featureID - the ID of the Feature
#
#  return:      a value that contains the translation of the 'featureID'
#               if the 'featureID' exists and it has a translation 
#               otherways the featureID
#
#________________________________________________________________________________
#
sub getFeatureTranslation {
    my $this = shift;
    my $result = undef;

    my $installuis   =$this->{installuis};
    my $uis1         =$this->{uis1};
    my $uis2         =$this->{uis2};
    my $jcluis       =$this->{jcluis};

    # check if the function has one parameter passed
    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter .\n");
    }

    my $featureID = shift;
    
    foreach my $anonimus (@$installuis) {
	if( $anonimus->{id} eq $featureID ) {
	    $result = $anonimus->{translation};
	    last;
	}
    }
    foreach my $anonimus (@$uis1) {
	if( $anonimus->{id} eq $featureID ) {
	    $result = $anonimus->{translation};
	    last;
	}
    }
    foreach my $anonimus (@$uis2) {
	if( $anonimus->{id} eq $featureID ) {
	    $result = $anonimus->{translation};
	    last;
	}
    }

    foreach my $anonimus (@$jcluis) {
	if( $anonimus->{id} eq $featureID ) {
	    $result = $anonimus->{translation};
	    last;
	}
    }

    if ( ! $result ) {
	$result = $featureID;
    }

    return $result;
}


#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  getFeatureDefault
#   
#  description: returns a value that contains the default option of a  feature 
#
#  parameters:  featureID - the ID of the Feature
#
#  return:      a value that contains the default option  of the 'featureID'
#               if the 'featureID' exists and it has an option 
#               otherways the 'undef' value
#
#________________________________________________________________________________
#
sub getFeatureDefault {
    my $this = shift;
    my $installuis   =$this->{installuis};
    my $uis1         =$this->{uis1};
    my $uis2         =$this->{uis2};
    my $jcluis       =$this->{jcluis};

    my $result = undef;

    # check if the function has one parameter passed
    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter .\n");
    }

    my $featureID = shift;
    
    foreach my $anonimus (@$installuis) {
	if( $anonimus->{id} eq $featureID ) {
	    $result = $anonimus->{default};
	    last;
	}
    }
    foreach my $anonimus (@$uis1) {
	if( $anonimus->{id} eq $featureID ) {
	    $result = $anonimus->{default};
	    last;
	}
    }
    foreach my $anonimus (@$uis2) {
	if( $anonimus->{id} eq $featureID ) {
	    $result = $anonimus->{default};
	    last;
	}
    }

    foreach my $anonimus (@$jcluis) {
	if( $anonimus->{id} eq $featureID ) {
	    $result = $anonimus->{default};
	    last;
	}
    }

    return $result;
}


#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  enumFeatureOptions
#   
#  description: returns a queue  that contains the options IDs of a feature 
#
#  parameters:  featureID - the ID of the Feature
#
#  return:      a queue that contains the options IDs of the feature 'featureID'
#               if the feature exists otherwaysan empty queue
#
#________________________________________________________________________________
#
sub enumFeatureOptions {
    my $this = shift;
    my $installuis   =$this->{installuis};
    my $uis1         =$this->{uis1};
    my $uis2         =$this->{uis2};
    my $jcluis       =$this->{jcluis};
    my @result = ();

    # check if the function has one parameter passed
    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter .\n");
    }

    my $featureID = shift;

    foreach my $anonimus (@$installuis) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		@result = ( @result , $option->{id} );
	    };
	    last;
	}
    }
    foreach my $anonimus (@$uis1) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		@result = ( @result , $option->{id} );
	    };
	    last;
	}
    }
    foreach my $anonimus (@$uis2) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		@result = ( @result , $option->{id} );
	    };
	    last;
	}
    }
    foreach my $anonimus (@$jcluis) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		@result = ( @result , $option->{id} );
	    };
	    last;
	}
    }
    
    return @result;
}


#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  setFeatureOptionTranslation
#   
#  description:  sets the translation of a feature option , of cource if the
#                feature and options exist
#                - is used to set the translation for features , options which
#                are not defined in the PPD file
#
#  parameters:  featureID - the ID of the Feature
#               optionID  - the ID of the Option
#               translation  - the translation of the optionID
#
#  return:      none
#________________________________________________________________________________
#
sub setFeatureOptionTranslation {
    my $this = shift;
    my $installuis   =$this->{installuis};
    my $uis1         =$this->{uis1};
    my $uis2         =$this->{uis2};
    my $jcluis       =$this->{jcluis};

    my $result = undef;

    # check if the function has three parameter passed
    if ( @_ < 3 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 3 parameter(s) .\n");
    }
    my $featureID = shift;
    my $optionID = shift;
    my $translation = shift;

    foreach my $anonimus (@$installuis) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		if ($option->{id} eq $optionID ) {
		    if( $option->{translation} ) {
			$option->{translation}=$translation;
			return;
		    }
		}
	    }
	}
    }
    foreach my $anonimus (@$uis1) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		if ($option->{id} eq $optionID ) {
		    if( $option->{translation} ) {
			$option->{translation}=$translation;
			return;
		    }
		}
	    }
	}
    }
    foreach my $anonimus (@$uis2) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		if ($option->{id} eq $optionID ) {
		    if( $option->{translation} ) {
			$option->{translation}=$translation;
			return;
		    }
		}
	    
	    }
	}
    }
    foreach my $anonimus (@$jcluis) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		if ($option->{id} eq $optionID ) {
		    if( $option->{translation} ) {
			$option->{translation}=$translation;
			return;
		    }
		}
	    
	    }
	}
    }
}



#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  getFeatureOptionTranslation
#   
#  description: returns a value that contains translation of a  feature  
#               option , of cource if there exists any translation 
#
#  parameters:  featureID - the ID of the Feature
#               optionID  - the ID of the Option
#
#  return:      a value that contains the translation of the 'optionID' of
#               the 'featureID' if the 'featureID' and 'optionID' exists and 
#               it has a translation otherways the 'undef' value
#
#________________________________________________________________________________
#
sub getFeatureOptionTranslation {
    my $this = shift;
    my $installuis   =$this->{installuis};
    my $uis1         =$this->{uis1};
    my $uis2         =$this->{uis2};
    my $jcluis       =$this->{jcluis};

    my $result = undef;

    # check if the function has one parameter passed
    if ( @_ < 2 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 2 parameter .\n");
    }
    my $featureID = shift;
    my $optionID = shift;

    foreach my $anonimus (@$installuis) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		if ($option->{id} eq $optionID ) {
		    if( $option->{translation} ) {
			$result=$option->{translation};
			last;
		    }
		}
		if ( $result ) {
		    last;
		}
	    }
	}
    }
    foreach my $anonimus (@$uis1) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		if ($option->{id} eq $optionID ) {
		    if( $option->{translation} ) {
			$result=$option->{translation};
			last;
		    }
		}
		if ( $result ) {
		    last;
		}
	    }
	}
    }
    foreach my $anonimus (@$uis2) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		if ($option->{id} eq $optionID ) {
		    if( $option->{translation} ) {
			$result=$option->{translation};
			last;
		    }
		}
		if ( $result ) {
		    last;
		}
	    }
	}
    }
    foreach my $anonimus (@$jcluis) {
	if( $anonimus->{id} eq $featureID ) {
	    my $options = $anonimus->{options};
	    foreach my $option (@$options) {
		if ($option->{id} eq $optionID ) {
		    if( $option->{translation} ) {
			$result=$option->{translation};
			last;
		    }
		}
		if ( $result ) {
		    last;
		}
	    }
	}
    }
    # if it has no translation use the ID
    if (! defined  $result ) {
	$result = $optionID;
    }
    return $result;
}

 
#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  get constraints hash
#   
#  description: returns a value that contains the ref. of the constraints hash
#
#  parameters: 
#
#  return:     return a ref. to the hash that contains the constraints ...
#
#________________________________________________________________________________
#
sub getConstraints {
    my $this = shift;
    my $constraints = $this->{constraints};
    return $constraints;
}
 

#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  isConstraint
#   
#  description:  check if the constraint exists
#
#  parameters: feat1 , opt1 ,feat2 , opt2
#
#  return:     0  - if the constraint does't exist
#              1  - if the constraint exists
#
#________________________________________________________________________________
#
sub checkConstraint {
    my $this = shift;
    my $constraints = $this->{constraints};

    # check if the function has one parameter passed
    if ( @_ < 4 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 4 parameter(s) .\n");
    }
    
    my $feat1 = shift;
    my $opt1  = shift;

    my $feat2 = shift;
    my $opt2  = shift;
    
    # eliminate from checking the *PageRegion
    if ($feat1 eq "*PageRegion" || $feat2 eq "*PageRegion" ) {
	return 0;
    }

    foreach my $constraintHash ( @$constraints ) {
	my $f1 = $constraintHash->{"f1"};
	my $o1 = $constraintHash->{"o1"};
	my $f2 = $constraintHash->{"f2"};
	my $o2 = $constraintHash->{"o2"};

	if( ( $f1 eq $feat1 && $o1 eq $opt1 && $f2 eq $feat2 && $o2 eq $opt2 ) ||
	    (  $f1 eq $feat2 && $o1 eq $opt2 && $f2 eq $feat1 && $o2 eq $opt1 ) ) {
	    return 1;

	}
    }
    return 0;

}

#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:  isCustomPaperSizeAvailable
#   
#  description:  check if the printer has Custom Page Size capabilities
#
#  parameters:  none
#
#  return:     0  - if is not available
#              1  - if is available
#
#________________________________________________________________________________
#
sub isCustomPaperSizeAvailable {
    my $this = shift;
    my $customPaperSize = $this->{"customPaperSize"};
    if ( defined $customPaperSize ) {
	return 1;
    } else {
	return 0;
    }
}


#________________________________________________________________________________
#
#  package:     Support::ibmXXXXX
#
#  subroutine:   getCustomPaperSizeParameter
#   
#  description:  return a parameter of the Custom Paper Size
#
#  parameters:   the name of the parameter :  "MinWidth"  - the min width
#                                             "MaxWidth"
#                                             "MinLength"
#                                             "MaxLength" 
#
#  return:    the value of the parameter
#             
#
#________________________________________________________________________________
#
sub getCustomPaperSizeParameter{
    my $this = shift;
    my $customPaperSize = $this->{"customPaperSize"};

    # check if the function has one parameter passed
    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter(s) .\n");
    }

    my $parameterKey = shift;

    if ( ! defined ( $customPaperSize) ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," : custom paper size not defined .\n",
	     "Check the existence with isCustomPaperSizeAvailable function\n" );
    }
    
    return $customPaperSize->{$parameterKey};
}




#################################################################################
#################################################################################
#
#                  package  Support::queues
#
#################################################################################
#################################################################################
package Support::queues;

use strict;
use vars qw( @ISA );

@ISA = qw( Support::base );


#________________________________________________________________________________
#
#  package:     Support::queues
#
#  subroutine:  new
#   
#  description: the constructor of the class   
#
#  parameters:  PAFfilename  - the full name  of the PAF file 
#
#  return:      the reference to the created object 
#
#________________________________________________________________________________
#	
sub new {
    my $class=shift;
    my $this=new Support::base( @_ );

    my @queues=();

    $this->{queues}=\@queues;

    bless($this,$class);

    # if the PAF file does'n exist create an empty PAF file ...
    if(! $this->isValid ) {
	$this->writeToPAF();
    }
    $this->fillFromPAF;
    return $this;   
}


#________________________________________________________________________________
#
#  package:     Support::queues
#
#  subroutine:  writeToPAF
#   
#  description: write data memebers to PAF file  
#
#  parameters:  ( no parameters )
#
#  return:      ( no value )
#________________________________________________________________________________
#
sub writeToPAF {
    my $this        = shift;
    my $PAFfilename = $this->{PAFfilename};
    my $queues      = $this->{queues};
    
    open (OUT_PAF , ">".$PAFfilename);
    # the first line is the control string
    print OUT_PAF ("## ",time,"\n");
    print OUT_PAF ("##\n");
    print OUT_PAF ("## this code is generated by IBMnpmenu\n");
    print OUT_PAF ("## please don\'t edit !!! \n");
    print OUT_PAF ("##\n");
    foreach my $anonimus (@$queues) {
	print OUT_PAF ("#queue\n");
	foreach   my $key    ( keys %$anonimus )  {
	    my $value = $anonimus->{$key};
	    if (! $value ) { 
		next;
	    }
	    if (ref($value) ne "HASH"){
		# if the value is not a hash write it
		print OUT_PAF ("# ".$key."\n");
		print OUT_PAF ($value."\n");
	    }
	    else {
		# if the value is a hash write the key and enter to the hash value
		print OUT_PAF ("#".$key."\n");
		# check all keys from the hash
		foreach   my $opt_key   ( keys %$value ) {
		    # get the value assoc to the key
		    my $opt_value = $value->{$opt_key};
		    if ( ref($opt_value) ne "HASH" ) {
			# if the value is a not a hash write it ...
			if( defined $opt_value) {
			    print OUT_PAF ("# ".$opt_key."\n");
			    print OUT_PAF ($opt_value."\n");
			}
		    } else {
			# the value is a hash
			# is custom paper sizes
			# write the custom paper size name
			print OUT_PAF ($opt_key."\n");
			# take all the options of the custom paper size and write them
			foreach my $sub_opt_key ( keys %$opt_value ) {
			    my $sub_opt_value = $opt_value->{$sub_opt_key};
			    print OUT_PAF ("# ".$sub_opt_key."\n");
			    print OUT_PAF ($sub_opt_value."\n");	    
			}
			print OUT_PAF ("#\$\$\n");
		    }
		}
		print OUT_PAF ("#\$\n");
	    }
	} 
	print OUT_PAF ("#end_queue\n");
    }
    close (OUT_PAF);
}




#________________________________________________________________________________
#
#  package:     Support::queues
#
#  subroutine:  fillFromPAF
#   
#  description: fill data members of the class form PAF 
#
#  parameters:  ( no parameters )
#
#  return:      ( no value )
#________________________________________________________________________________
#
sub fillFromPAF {
    my $this = shift;
    my $PAFfilename = $this->{PAFfilename};
    my $queues = $this->{queues};
    @$queues = (); 
    my $currentQueue = {};
    my $currentOpt   = {};
    my $currentCustomPaperSizes = {};
    my $paperSize= {};
    my $key = undef;
    my $value;

    my @buffer;
    my $status = "undef" ; # it can be "param" or "options" or "undef"
    open ( IN_PAF , $PAFfilename );
    @buffer=<IN_PAF>;
    close( IN_PAF );

    # start to parse the buffer ...
    chop(@buffer);

    foreach    my $item  (@buffer) {
        # out with blank lines 
	if ( $item eq '' || substr($item,0,1) eq ' ') {
	    next;
	}
        # out with comments 
	if ( substr($item,0,2) eq '##' ) {
	    next;
	}
        # start to parse queues
	if ( substr($item,0,6) eq '#queue' ) {       # start a new queue
	    $status = "param";
            # get ready for start a new queue
	    $currentQueue  = {};
	    next;
	}
	if ( substr($item,0,10) eq '#end_queue' ) {  # end the curent queue
	    $status = "param";
            # get ready to save the queue
	    @$queues = ( @$queues , $currentQueue );
	    next;
	}
	if ( substr($item,0,8) eq '#options' ) {    # start options section
	    $status = "options";
            # get ready to start options
	    $currentOpt={};
            # get ready to save  options 
	    $currentQueue->{options}=$currentOpt;
	    next;
	}
	if (substr($item,0,17) eq "#customPaperSizes" ) {
	    $status="customPaperSizes";
            # get ready to start custom paper sizes
	    $currentCustomPaperSizes={};
	    # get ready to save the custom paper sizes
	    $currentQueue->{"customPaperSizes"}=$currentCustomPaperSizes;
	    next;
	}
	if ( substr($item,0,3) eq '#$$' ) {
	    # end 'paperSize' and return to 'customPaperSizes'
	    $status="customPaperSizes";
	    next;
	}
	if ( substr($item,0,2) eq '#$' ) {          # end options section'
	    $status = "param";
	    next;
	}       

       if (!  $key  ) {
	   if($status eq 'customPaperSizes') {
	       $paperSize={};
	       $currentCustomPaperSizes->{$item}=$paperSize;
	       $status="paperSize";
	   } else {
	       $key = substr($item,2);
	   }
           next;
        }

        $value=$item;
        if($status eq 'param') {
           $currentQueue->{$key}=$value;
           $key = undef;
        }
        elsif ($status eq 'options' ) {
          $currentOpt->{$key}=$value;
	  $key = undef;
      } elsif ( $status eq 'paperSize') {
	  $paperSize->{$key}=$value;
	  $key = undef;
      }
	
 
    }  
    
}#________________________________________________________________________________
#
#  subroutine:  enumQueues
#   
#  description: returns an array with the Queue names
#
#  parameters:  ( no parametters )
#
#  return:     an array with all queues   if there exists any queue
#               otherways  a blank array .
#________________________________________________________________________________
#
sub enumQueues {
    my $this = shift;
    my $queues = $this->{queues};
    my @result = ();

    foreach my $anonimus (@$queues) {
	@result = ( @result , $anonimus->{queueName} );
    }
    return @result;
}

#________________________________________________________________________________
#
#  subroutine:  deleteQueue
#   
#  description: deletes a queue from queues list 
#
#  parameters:  queueName - the name of the queue which will be deleted
#
#  return:     ( no value )
#
#________________________________________________________________________________
#
sub deleteQueue {
    my $this = shift;
    my $queues = $this->{queues};
    my @new_queues = ();

    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter(s) .\n");
    }

    my $queueName = shift;

    foreach my $anonimus (@$queues) {
	if (  $anonimus and $anonimus->{queueName} ne $queueName ) {
	    @new_queues =( @new_queues , $anonimus );
	}
    }
    $this->{queues} = \@new_queues;
}
#________________________________________________________________________________
#
#  subroutine:  changeQueueName
#   
#  description: change the queuename 
#
#  parameters:  queueName - the old queueName
#               newQueueName - the new queueName
#
#  return:     ( no value )
#
#________________________________________________________________________________
#
sub changeQueueName {
    my $this = shift;
    my $queues = $this->{queues};

    if ( @_ < 2 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 2 parameter(s) .\n");
    }

    my $queueName    = shift;
    my $newQueueName = shift;

    foreach my $anonimus (@$queues) {
	if (  $anonimus->{queueName} eq $queueName ) {
	    $anonimus->{queueName} = $newQueueName;
	}
    }
}

#________________________________________________________________________________
#
#  subroutine:  getQueueProperty
#   
#  description: returns a property of a queue
#
#  parameters:  queueName - the name of the Queue
#               property  - the name of the property searched
#                    available properties :
#                       "printer" - the name of the printer
#                       "PPD"     - the PPD file of the printer
#                       "type"    - the transport type
#                       "address" - the IP address of the printer
#                                   is valid only if "type" is "TCP/IP"
#                       "port"    - the port of the printer
#                                   is valid only if "type" is "Serial" or 
#                                                                    "Parallel"
#                       "language" - is the lang. of the job : Pcl , PS , Raw
#
#  return:      the 'property' value for the 'queueName' if exists
#               otherways  'undef'  value .
#________________________________________________________________________________
#
sub getQueueProperty {
    my $this = shift;
    my $queues = $this->{queues};
    my $result = undef;

    if ( @_ < 2 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 2 parameter .\n");
    }
    my $searchKey = shift;
    my $property  = shift;

    foreach my $anonimus (@$queues) {
	if ( $anonimus->{queueName} eq $searchKey ) {
	    $result = $anonimus->{$property};
	    last;
	}
    }
    return $result;
} 
#________________________________________________________________________________
#
#  subroutine:  setQueueProperty
#   
#  description: sets the propertyes  of a queue
#
#  parameters:  queueName - the name of the Queue
#               property  - the name of the property searched
#                    available properties :
#                       "printer" - the name of the printer
#                       "PPD"     - the PPD file of the printer
#                       "type"    - the transport type
#                       "address" - the IP address of the printer
#                                   is valid only if "type" is "TCP/IP"
#                       "port"    - the port of the printer
#                                   is valid only if "type" is "Serial" or 
#                                                                    "Parallel"
#                       "language" - is the lang. of the job : Pcl , PS , Raw
#
#               propertyValue - the value of the property
#
#  return:      ( no value ) 
#________________________________________________________________________________
#
sub setQueueProperty {
    my $this = shift;
    my $queues = $this->{queues};

    if ( @_ < 3 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 2 parameter .\n");
    }
    my $searchKey = shift;
    my $property  = shift;
    my $propertyValue = shift;

    foreach my $anonimus (@$queues) {
	if ( $anonimus->{queueName} eq $searchKey ) {
	    $anonimus->{$property} = $propertyValue;
	    last;
	}
    }
} 
#________________________________________________________________________________
#
#  subroutine:  getQueueOptions
#   
#  description: returns a hash with all options ( different from default ) 
#               of the printer
#               associated with a queue
#
#  parameters:  queueName - the name of the Queue
#
#  return:      the hash with all options ( different from default ) 
#               of the printer associated the 'queueName' if exists 
#               otherways  an empty  hash  .
#________________________________________________________________________________
#
sub getQueueOptions {
    my $this = shift;
    my $queues = $this->{queues};
    my %result = ();

    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 2 parameter .\n");
    }
    my $searchKey = shift;

    foreach my $anonimus (@$queues) {
	if ( $anonimus->{queueName} eq $searchKey ) {
	    %result = %{$anonimus->{options}}; 
	    last;
	}
    }

    return \%result;
} 
#________________________________________________________________________________
#
#  subroutine:  setQueueOptions
#   
#  description: add  all options ( different from default ) 
#               of the printer associated with a queue
#
#  parameters:  queueName   - the name of the Queue
#               hashOptions - the options
#
#  return:      ( no value )
#________________________________________________________________________________
#
sub setQueueOptions {
    my $this = shift;
    my $queues = $this->{queues};

    if ( @_ < 2 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 2 parameter .\n");
    }
    my $searchKey   = shift;
    my $hashOptions = shift;

    # erase all keys which has undef values
    my $newHash = {};
    foreach my $key ( keys %$hashOptions ) {
	if ( $hashOptions->{$key}) {
	    $newHash->{$key} = $hashOptions->{$key};
	}
    }

    foreach my $anonimus (@$queues) {
	if ( $anonimus->{queueName} eq $searchKey ) {
	    $anonimus->{options}=$newHash; 
	    last;
	}
    }
} 

#________________________________________________________________________________
#
#  subroutine:  getQueueCustomPaperSizes
#   
#  description: returns a hash with all custom paper sizes defined
#               for that queue , if the queue is assoc. to a printer
#               that supports custom paper sizes.
#
#  parameters:  queueName - the name of the Queue
#
#  return:     undef - if  the queue does not support custom paper sizes
#              a ref. to a hash with the custom paper sizes
#________________________________________________________________________________
#
sub getQueueCustomPaperSizes {
    my $this = shift;
    my $queues = $this->{"queues"};

    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter(s) .\n");
    }
    my $searchKey = shift;

    foreach my $anonimus (@$queues) {
	if ( $anonimus->{queueName} eq $searchKey ) {
	    return $anonimus->{"customPaperSizes"}; 
	}
    }

    return undef;
} 
#________________________________________________________________________________
#
#  subroutine:  setQueueCustomPaperSizes
#   
#  description: add the hash with custom paper sizes to the queue
#
#  parameters:  queueName         - the name of the Queue
#               customPaperSizes  - the hash with custom paper sizes
#
#  return:      ( no value )
#________________________________________________________________________________
#
sub setQueueCustomPaperSizes {
    my $this = shift;
    my $queues = $this->{queues};

    if ( @_ < 2 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 2 parameter(s) .\n");
    }


    my $searchKey   = shift;
    my $hash        = shift;

    if( ! defined  $hash) {
	return;
    }

    # erase all keys which has undef values
    my $newHash = {};
    foreach my $key ( keys %$hash ) {
	if ( $hash->{$key}) {
	    $newHash->{$key} = $hash->{$key};
	}
    }

    foreach my $anonimus (@$queues) {
	if ( $anonimus->{queueName} eq $searchKey ) {
	    $anonimus->{"customPaperSizes"}=$newHash; 
	    last;
	}
    }
} 


#________________________________________________________________________________
#
#  subroutine:  addDuplicatedQueue
#   
#  description: add a queue which has the same data as the parameter queue
#   
#  parameters:  queueName - the name of the queue which is cloned
#               clonedQueueName - the name of the cloned queue
#
#  return:      ( vo value )
#________________________________________________________________________________
#
sub addDuplicateQueue {
    my $this = shift;
    my $queues = $this->{queues};
    if ( @_ < 2 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter(s) .\n");
    }
    my $queueName   = shift;
    my $clonedQueueName = shift;

    my $anonimus = {};
    $anonimus->{"queueName"}=$clonedQueueName;
    @$queues = ( @$queues , $anonimus );
    if ( $queueName ) {
	$this->setQueueProperty($clonedQueueName,"printer",
				$this->getQueueProperty($queueName,"printer"));
	$this->setQueueProperty($clonedQueueName,"PPD",
				$this->getQueueProperty($queueName,"PPD"));
	$this->setQueueProperty($clonedQueueName,"type",
				$this->getQueueProperty($queueName,"type"));
	$this->setQueueProperty($clonedQueueName,"address",
				$this->getQueueProperty($queueName,"address"));
	$this->setQueueProperty($clonedQueueName,"port",
				$this->getQueueProperty($queueName,"port"));
	$this->setQueueProperty($clonedQueueName,"language",
				$this->getQueueProperty($queueName,"language"));
	$this->setQueueOptions($clonedQueueName,
			       $this->getQueueOptions($queueName));
	if(defined $this->getQueueCustomPaperSizes($queueName)) {
	    $this->setQueueCustomPaperSizes($clonedQueueName,
					    $this->getQueueCustomPaperSizes($queueName));
	}
    }
    else {
	$anonimus->{options}={};
	$anonimus->{customPaperSizes}={};
    }
}

#________________________________________________________________________________
#
#  subroutine:  addClonedQueue
#   
#  description: add a cloned queue in order to be edited  
#   
#  parameters:  queueName - the name of the queue which is cloned
#               clonedQueueName - the name of the cloned queue
#
#  return:      ( vo value )
#________________________________________________________________________________
#
sub addClonedQueue {
    my $this = shift;
    my $queues = $this->{queues};
    if ( @_ < 2 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter(s) .\n");
    }
    my $queueName   = shift;
    my $clonedQueueName = shift;

    my $anonimus = {};
    $anonimus->{"queueName"}=$clonedQueueName;
    $anonimus->{"editedQueueName"}=$queueName;
    $anonimus->{"queueType"}="edit";
    @$queues = ( @$queues , $anonimus );
    if ( $queueName ) {
	$this->setQueueProperty($clonedQueueName,"printer",
				$this->getQueueProperty($queueName,"printer"));
	$this->setQueueProperty($clonedQueueName,"PPD",
				$this->getQueueProperty($queueName,"PPD"));
	$this->setQueueProperty($clonedQueueName,"type",
				$this->getQueueProperty($queueName,"type"));
	$this->setQueueProperty($clonedQueueName,"address",
				$this->getQueueProperty($queueName,"address"));
	$this->setQueueProperty($clonedQueueName,"port",
				$this->getQueueProperty($queueName,"port"));
	$this->setQueueProperty($clonedQueueName,"language",
				$this->getQueueProperty($queueName,"language"));
	$this->setQueueOptions($clonedQueueName,
			       $this->getQueueOptions($queueName));
	if(defined $this->getQueueCustomPaperSizes($queueName)) {
	    $this->setQueueCustomPaperSizes($clonedQueueName,
					    $this->getQueueCustomPaperSizes($queueName));
	}
    }
    else {
	$anonimus->{options}={};
	$anonimus->{customPaperSizes}={};
    }
}
#________________________________________________________________________________
#
#  subroutine:  IsCloned
#   
#  description: verify if the queue is a cloned queue  
#   
#  parameters:  queueName - the name of the queue
#
#  return:      1 - if the queue is a clone
#               0 - if the queue is not a clone
#________________________________________________________________________________
#
sub isCloned {
    my $this = shift;
    my $queues = $this->{queues};
    my $result = 0;
    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter(s) .\n");
    }
    my $queueName   = shift;
    foreach my $anonimus (@$queues) {
	if ( $anonimus->{queueName} eq $queueName and 
	     $anonimus->{queueType} eq "edit" ) {
	    $result = 1;
	    last;
	}
    }
    return $result;
}
#________________________________________________________________________________
#
#  subroutine:  IsQueue
#   
#  description: verify if the queue esist in the list  
#   
#  parameters:  queueName - the name of the queue
#
#  return:      1 - if the queue exists
#               0 - if the queue doesn't exist
#________________________________________________________________________________
#
sub isQueue {
    my $this = shift;
    my $queues = $this->{queues};
    my $result = 0;
    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter(s) .\n");
    }
    my $queueName   = shift;
    foreach my $anonimus (@$queues) {
	if ( $anonimus->{queueName} eq $queueName ) {
	    $result = 1;
	    last;
	}
    }
    return $result;
}


#________________________________________________________________________________
#
#  subroutine:  checkConstraints
#   
#  description: check if the constranis are not violated   
#   
#  parameters:  queueName - the name of the queue
#               printer   - the printer assoc. to the queue
#
#  return:      a hash with the invalid combination
#________________________________________________________________________________
#
sub checkConstraints {
    my $this   = shift;
    my $queues = $this->{queues};

    my %allOptionsSetted=();
    my $result = undef;

    if ( @_ < 2 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter(s) .\n");
    }
    my $queueName   = shift;    
    my $printer     = shift;

    # check if the queue exists
    if(! $this->isQueue($queueName)) {
	return $result;
    }

    # get the custom
    my $pageSize;
    my $cpsHash= $this->getQueueCustomPaperSizes($queueName);

    # get the queue Options
    my $queueOptions=undef;
    $queueOptions=$this->getQueueOptions($queueName);
    # copy into a new location the options hash
    %allOptionsSetted=%$queueOptions;
 
    # complete the new option hash with the  other options with the
    # default position
    # now add the default setting for the options not setted
    foreach my $printerFeatureName ( $printer->enumFeatures() ){
	if ( ! $allOptionsSetted{$printerFeatureName} ) {
	    $allOptionsSetted{$printerFeatureName}=$printer->getFeatureDefault($printerFeatureName);
	}
    } 
    
 
    # in %allOptionSetted if the *PageSize is a custom paper size
    # remove the *PaperSize and insert *CustomPageSize True
    $pageSize=$allOptionsSetted{"*PageSize"};
    if( defined $cpsHash && 
	defined $cpsHash->{$pageSize}) {
	# a custom page size had been selected
	$allOptionsSetted{"*PageSize"}=undef;
	$allOptionsSetted{"*CustomPageSize"}="True";
    }

    # now I have in the hash : %allOptionsSetted all features with
    #  selected ( or default ) options
    # now it's SHOW TIME !!! -- check the constraints
    foreach my $feat1 ( keys %allOptionsSetted ) {
	my $opt1=$allOptionsSetted{$feat1};
	foreach my $feat2 ( keys %allOptionsSetted ) {
	    my $opt2 = $allOptionsSetted{$feat2};
	    #
	    if(defined $opt1 && defined $opt2 &&
	       $feat1 ne $feat2 &&
	       $printer->checkConstraint($feat1,$opt1,$feat2,$opt2)) {
		$result->{f1}=$feat1;
		$result->{o1}=$opt1;
		$result->{f2}=$feat2;
		$result->{o2}=$opt2;
		return $result;
	    }
	    #
	}
    }
    return  $result;
}


#________________________________________________________________________________
#
#  subroutine:  isVisualConstraint
#   
#  description: check if the constranis are not violated   
#   
#  parameters:  queueName - the name of the queue
#               printer   - the printer assoc. to the queue
#               feature
#               option
#
#  return:      1 - if there is a constraint violated
#               0 - no constraint was violated
#________________________________________________________________________________
#
sub isVisualConstraint {
    my $this   = shift;
    my $queues = $this->{"queues"};
    my %allOptionsSetted=();
   
    my $result = 0;

    if ( @_ < 4 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 4 parameter(s) .\n");
    }
    my $queueName   = shift;    
    my $printer     = shift;
    my $feature     = shift;
    my $option      = shift;

    # check if the queue exists
    if(! $this->isQueue($queueName)) {
	return $result;
    }

    # get the custom
    my $pageSize;
    my $cpsHash= $this->getQueueCustomPaperSizes($queueName);

    # check if feature,option is *PageSize,a custom page defined
    $pageSize=$option;
    if( $feature eq "*PageSize" &&
	defined $cpsHash && 
	defined $cpsHash->{$pageSize}) {
	    $feature="*CustomPageSize";
	    $option="True";
	}

    # get the queue Options
    my $queueOptions=undef;
    $queueOptions=$this->getQueueOptions($queueName);
    # copy into a new location the options hash
    %allOptionsSetted=%$queueOptions;
 
    # complete the new option hash with the  other options with the
    # default position
    # now add the default setting for the options not setted
    foreach my $printerFeatureName ( $printer->enumFeatures() ){
	if ( ! $allOptionsSetted{$printerFeatureName} ) {
	    $allOptionsSetted{$printerFeatureName}=$printer->getFeatureDefault($printerFeatureName);
	}
    } 
    
    # in %allOptionSetted if the *PageSize is a custom paper size
    # remove the *PaperSize and insert *CustomPageSize True
    $pageSize=$allOptionsSetted{"*PageSize"};
    if( defined $cpsHash && 
	defined $cpsHash->{$pageSize}) {
	# a custom page size had been selected
	$allOptionsSetted{"*PageSize"}=undef;
	$allOptionsSetted{"*CustomPageSize"}="True";
    }

    # now I have in the hash : %allOptionsSetted all features with
    #  selected ( or default ) options
    # now it's SHOW TIME !!! -- check the constraints
    foreach my $feat1 ( keys %allOptionsSetted ) {
	my $opt1=$allOptionsSetted{$feat1};
	    if(defined $opt1 && 
	       $feat1 ne $feature &&
	       $printer->checkConstraint($feature,$option,$feat1,$opt1)) {
		$result = 1;
		return $result;
	    }
	
    }

    return  $result;

}

#________________________________________________________________________________
#
#  subroutine:  getClonedQueueName
#   
#  description: get the real name of the cloned queue  
#   
#  parameters:  clonedQueueName - the name of the cloned queue
#
#  return:      the real name of the cloned queue
#________________________________________________________________________________
#
sub getClonedQueueName {
    my $this   = shift;
    my $queues = $this->{queues};
    my $result = undef;
    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter(s) .\n");
    }
    my $clonedQueueName   = shift;
    foreach my $anonimus (@$queues) {
	if ( $anonimus->{queueName} eq $clonedQueueName and 
	     $anonimus->{queueType} eq "edit" ) {
	    $result = $anonimus->{editedQueueName};
	    last;
	}
    }

    return $result;
}

#________________________________________________________________________________
#
#  subroutine:  cleanQueueGarbage
#   
#  description: clean the queue informations added during edit ...  
#   
#  parameters:  queueName - the name of the  queue
#
#  return:      ( no value )
#________________________________________________________________________________
#
sub cleanQueueGarbage  {
    my $this   = shift;
    my $queues = $this->{queues};
    my $result = undef;
    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter(s) .\n");
    }
    my $queueName   = shift;
    foreach my $anonimus (@$queues) {
	if ( $anonimus->{queueName} eq $queueName ) {
	    $anonimus->{editedQueueName} = undef;
	    $anonimus->{type} = undef;
	    last;
	}
    }
    return $result;
}


#________________________________________________________________________________
#
#  subroutine:  writePrintcapFile
#   
#  description: update the printcap file  
#   
#  parameters:  printcapFile - the full path to printcap file
#               filterExe    - the full path of the filter  
#
#  return:      the error which ocured
#               'undef' if there is no error  
#________________________________________________________________________________
#
sub writePrintcapFile  {
    my $this   = shift;
    my $queues = $this->{queues};
    my $introduction1 = "# the following lines are inserted by IBMnpmenu";
    my $introduction2 = "# please do not edit";
    my $introduction3 = "#";
    my $endQueueIBMnpmenu = ':if=';
    if ( @_ < 3 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 3 parameter(s) .\n");
    }

    my $printcapFile   = shift;
    my $filterExe      = shift;
    my $datFolder      = shift;
    my @buffer = ();
    my @outBuffer = ();
#  open printcap file and read all data from it ...
    if( -e $printcapFile) {
	open (PRC , $printcapFile);
	@buffer = <PRC>;
	close(PRC);
    }
# clean the buffer ...
    my $semaphore = 0;
    foreach $_ (@buffer) {
	if(m/$introduction1/) {
	    $semaphore = 1;
	    next;
	}
	if( $semaphore and m/$endQueueIBMnpmenu/ ) {
	    $semaphore = 0;
	    next;
	}
	if (! $semaphore ) {
	    @outBuffer = ( @outBuffer , $_ );
	}
    }
# continue the outBuffer with queues defined by IBMnpmenu ... 

# write data in the printcap file 
    foreach my $anonimus (@$queues) {
	@outBuffer = ( @outBuffer , $introduction1                ."\n");
	@outBuffer = ( @outBuffer , $introduction2                ."\n");
	@outBuffer = ( @outBuffer , $introduction3                ."\n");
	@outBuffer = ( @outBuffer , $anonimus->{queueName}.":\\"  ."\n");
	@outBuffer = ( @outBuffer , "\t:force_localhost\@:\\"     ."\n");
	@outBuffer = ( @outBuffer , "\t:try_localhost\@:\\"       ."\n");
	@outBuffer = ( @outBuffer , "\t:mx#0:\\"                  ."\n");
	@outBuffer = ( @outBuffer , "\t:sh:\\"                    ."\n");
	@outBuffer = ( @outBuffer , "\t:sd=/var/spool/lpd/$anonimus->{queueName}:\\"  ."\n");
	# also create the spool folder
	if( -e "/var/spool/lpd/$anonimus->{queueName}") {
	} else {
	    system("mkdir /var/spool/lpd/$anonimus->{queueName}");
	}
	@outBuffer = ( @outBuffer , "\t:lpr_bounce:\\"            ."\n");
	my $printerPortAddress="";
	$printerPortAddress.=":lp=";
# if the printer is TCP/IP i have to write :lp=pppp@aaaaaaaaaaaa:\ 
#                   serial or parallel     :lp=xxxxx:\
	if ($anonimus->{type} eq "TCP/IP" ) {
	    $printerPortAddress.="PASS"."@";
	    $printerPortAddress.=$anonimus->{address};
	}
	else {
	    $printerPortAddress.=$anonimus->{port};
	}
	@outBuffer = (@outBuffer , "\t".$printerPortAddress.":");

	if( $anonimus->{language} eq  "PS" ) {
# setup the filter 
	    @outBuffer = (@outBuffer , "\\\n");
	    my $filterString="";
	    $filterString.=":if=";
	    $filterString.=$filterExe;
	    $filterString.=" --ppdfile ";
	    $filterString.=$datFolder."/"."ppds"."/".$anonimus->{PPD};
	    $filterString.=" --deffile ";
	    $filterString.=$datFolder."/".$anonimus->{queueName}.".def";
	    $filterString.=" --filter ";
	    $filterString.=":";
	    @outBuffer = (@outBuffer , "\t".$filterString            ."\n");
	}
    }
    open (PRC , ">".$printcapFile);
    print PRC @outBuffer;
    close(PRC);
    

}


#________________________________________________________________________________
#
#  subroutine:  writeDefFile
#   
#  description: write the .def file for the queue  
#   
#  parameters:  pathToDef
#
#________________________________________________________________________________
#
sub writeDefFiles  {

    my $this   = shift;
    my $queues = $this->{queues};
    if ( @_ < 1 ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :function requires 1 parameter(s) .\n");
    }

    my $path   = shift;
    foreach my $anonimus (@$queues) {
	my $defFilename = $path."/".$anonimus->{queueName}.".def";
	my $options = $anonimus->{options};
	open ( DEF , ">".$defFilename);
	foreach my $key  ( keys %$options ) {
	    if ( $options->{$key} ) {
		print DEF "$key=".$options->{$key}."\n";
	    }
	}
	my $queueName = $anonimus->{queueName};
	my $customPaperSizes=$this->getQueueCustomPaperSizes($queueName);
	foreach my $key ( keys %$customPaperSizes) {
	    my $customPageHash=$customPaperSizes->{$key};
	    my $width  = $customPageHash->{"Width"};
	    my $length = $customPageHash->{"Length"};
	    print DEF "*CustomPaperSize=$key\'\'$width\'\'$length\n";
	}
	close ( DEF );
    }
}




#################################################################################
#################################################################################
#
#                           package  Support::printers
#
#################################################################################
#################################################################################
package Support::printers;

use strict;

#________________________________________________________________________________
#
#  package:     Support::printers
#
#  subroutine:  new
#   
#  description: the constructor of the class   
#
#  parameters:  the PPD folder , the PAF folder
#
#  return:      the reference to the created object 
#
#________________________________________________________________________________
#	
sub new {

    my $class = shift;
    my $this  = {};
    
    if(  @_ >= 2) {
	$this->{ PPDfolder } = shift;
	$this->{ PAFfolder } = shift;
    }
    else {
	die " Error !!! $class : PPD Folder and PAF Folder  required \n ";
    }

    my %printers=();
    $this->{ printers } =\%printers;
 
    bless( $this , $class );


    $this->discoverPrinters();

    return $this;

}

#________________________________________________________________________________
#
#  package:     Support::printers
#
#  subroutine:   checkPAFfile
#   
#  description:  check if paf file is valid file
#
#  parameters:  PAFfilename
#               PPDfilename
#
#  return:      1    - PAF is a valid file
#               0    - PAF is not valid 
#________________________________________________________________________________
#
sub checkPAFfile {
    my $this = shift;
    my $PAFfolder = $this->{PAFfolder};
    my $PPDfolder = $this->{PPDfolder};

    my $PAFfilename=undef;
    my $PPDfilename=undef;

    if(  @_ >= 2) {
	$PAFfilename = shift;
	$PPDfilename = shift; 
    }
    else {
	die " Error !!! checkPAFfile:checkPAFfile -  PAF filename and  PPD filename  required  \n";
    }

    # check for it's existence 
    if(! -e $PAFfolder."/".$PAFfilename ) {
	return 0;
    }

    # if exists check if is a PAF file
    
    open ( PAF ,  $PAFfolder."/".$PAFfilename );
    my $buffer = undef ;
    $buffer=<PAF>;$buffer=<PAF>;$buffer=<PAF>;
    $buffer=<PAF>;$buffer=<PAF>;$buffer=<PAF>;
    $buffer=<PAF>;
    chop $buffer;
    close( PAF );
    my $extractedPPDfilename=substr($buffer,8);
    if(  $extractedPPDfilename eq $PPDfilename ) {
	return 1;
    } 

    return 0;
}


#________________________________________________________________________________
#
#  package:     Support::printers
#
#  subroutine:   registerPrinter
#   
#  description:  register the printer
#
#  parameters:  PAFfilename
#               PPDfilename
#
#  return:     ( no value )
#________________________________________________________________________________
#
sub registerPrinter {
    my $this = shift;
    my $PAFfolder = $this->{PAFfolder};
    my $PPDfolder = $this->{PPDfolder};
    my $printers  = $this->{printers};
    my $paffile = undef; 
    my $ppdfile = undef; 

    if(  @_ >= 2) {
	$paffile = shift;
	$ppdfile = shift;
    }
    else {
	die " Error !!! printers::translate2PAF -  PAF filename and  PPD filename  required  \n";
    }
    open ( PAF ,  $PAFfolder."/".$paffile );
    my $buffer = undef ;
    $buffer=<PAF>;$buffer=<PAF>;$buffer=<PAF>;
    $buffer=<PAF>;$buffer=<PAF>;
    $buffer=<PAF>;
    chop $buffer;
    close( PAF );
    my $printerName=substr($buffer,9);
    my %printer=();
    $printer{ppd}=$ppdfile;
    $printer{paf}=$paffile;
    $printers->{$printerName}=\%printer;
}

#________________________________________________________________________________
#
#  package:     Support::printers
#
#  subroutine:   translate2PAF
#   
#  description:  translate from PPD to PAF files
#
#  parameters:  PAF filename
#               PPD filename
#
#  return:      1    - PAF is a valid file ( was created OK )
#               0    - PAF is not valid    ( error during creation )
#________________________________________________________________________________
#
sub translate2PAF {
    my $this = shift;
    my $PAFfolder = $this->{PAFfolder};
    my $PPDfolder = $this->{PPDfolder};
    my $paffile = undef; 
    my $ppdfile = undef; 

    if(  @_ >= 2) {
	$paffile = shift;
	$ppdfile = shift;
    }
    else {
	die " Error !!! printers::translate2PAF -  PAF filename and  PPD filename  required  \n";
    }

    print "Building $paffile from $ppdfile  ....\n";
    my $ppdfullname = $PPDfolder."/".$ppdfile;
    my $paffullname = $PAFfolder."/".$paffile;
    system("PPDParser --ppdfile $ppdfullname --paffile $paffullname 2>&1 &> /dev/null");
    if(!$this->checkPAFfile($paffile,$ppdfile)) {
	print "Failed to build  $paffile \n\n";
    }
    else {
	print "$paffile was successfuly build\n\n";
	$this->registerPrinter($paffile,$ppdfile);
    }

}


#________________________________________________________________________________
#
#  package:     Support::printers
#
#  subroutine:  count supported printers
#   
#  description: search for the supported printers 
#
#  parameters:  ( no parameter )
#
#  return:      ( no value )
#________________________________________________________________________________
#
sub discoverPrinters  {
    my $this = shift;
    my $PAFfolder = $this->{PAFfolder};
    my $PPDfolder = $this->{PPDfolder};

    my @PPDfiles  = ();
    my %PPD2PAFfiles = ();
    my %needTranslation=();

    # search for ppd files in the PPDFolder
    if(! -d $PPDfolder ) {
	die ("Error  :file ",__FILE__,":line ",
	     __LINE__," :  $PPDfolder is an invalid folder \n");
    }
    
    # make the list of PPD files
    opendir( PPD_DIR , $PPDfolder );
    while( my $ppdFile=readdir(PPD_DIR) ) {
	my $error=0;
	if( -f $PPDfolder."/".$ppdFile) {
	    # test the extension of the file
	    my  $test=substr($ppdFile,-3);
	    if($test ne "ppd") {
		$error=1;
	    }
	    #
	    # Here I can test also if the first non-empty line is the PPD signature
	    #
	    
	    # add the ppd file if is valid to the list
	    if( $error == 0 ) {
		@PPDfiles=(@PPDfiles,$ppdFile);
	    }
	}
    }
    closedir( PPD_DIR );



    # if PAFFolder exists then check if PAF files are OK !!!
    # process PAF filenames from PPD files
    foreach my $ppdFile ( @PPDfiles ) {
	my $pafFile = $ppdFile;
	$pafFile =~ s/\.ppd/\.paf/;
	$PPD2PAFfiles{$ppdFile}=$pafFile;
    }

    # check if PAFFoder exists 
    # search for ppd files in the PPDFolder
    if(! -d $PAFfolder ) {
	# if it doesn't exist create it and 
	# call PAF generator function for all available PPDs.
	system("mkdir $PAFfolder > /dev/null");
	foreach my $ppdfile ( keys %PPD2PAFfiles ) {
	    $this->translate2PAF($PPD2PAFfiles{ppdfile} , $ppdfile );
	}
	return;
    }


    # check the validity of the PAF file 
    foreach my $ppdFile ( keys %PPD2PAFfiles ) {
	if($this->checkPAFfile( $PPD2PAFfiles{$ppdFile},$ppdFile)) {
	    # if the PAF file is OK add it to the list of PAFs
	    $this->registerPrinter($PPD2PAFfiles{$ppdFile},$ppdFile);
	}
	else {
	    $needTranslation{$ppdFile}=$PPD2PAFfiles{$ppdFile};
	}
    }

    # try to build the unexistent PAF files ...
    foreach my $ppdfile ( keys %needTranslation) {
	    print "$needTranslation{$ppdfile} invalid file. \n";
	$this->translate2PAF($needTranslation{$ppdfile} , $ppdfile );
    }
}




#________________________________________________________________________________
#
#  package:     Support::printers
#
#  subroutine:  enumPrinters
#   
#  description: return a list of the supported printers  
#
#  parameters:  ( no parameter )
#
#  return:      a list of printer names
#________________________________________________________________________________
#
sub enumPrinters {

    my $this = shift;
    my $PAFfolder = $this->{PAFfolder};
    my $PPDfolder = $this->{PPDfolder};
    my $printers  = $this->{printers};

    my @printernamesList = ();

    @printernamesList = keys %$printers;

    return @printernamesList;
}


#________________________________________________________________________________
#
#  package:     Support::printers
#
#  subroutine:  getPrinterPPDfilename
#   
#  description: returns the PPD filename assoc. to the printer 
#
#  parameters:  the printer Name 
#
#  return:      the PPD filename
#________________________________________________________________________________
#
sub getPrinterPPDfilename {

    my $this = shift;
    my $PAFfolder = $this->{PAFfolder};
    my $PPDfolder = $this->{PPDfolder};
    my $printers  = $this->{printers};

    my $printerName = undef;

    if(  @_ >= 1) {
	$printerName = shift;
    }
    else {
	die " Error !!! printers::getPrinterPPDfilename - printer name required  \n";
    }


    my $printer = $printers->{$printerName} ;

    return $printer->{ppd};
}

#________________________________________________________________________________
#
#  package:     Support::printers
#
#  subroutine:  getPrinterPPDfullname
#   
#  description: returns the PPD filename assoc. to the printer 
#
#  parameters:  the printer Name 
#
#  return:      the PPD filename
#________________________________________________________________________________
#
sub getPrinterPPDfullname {

    my $this = shift;
    my $PAFfolder = $this->{PAFfolder};
    my $PPDfolder = $this->{PPDfolder};
    my $printers  = $this->{printers};

    my $printerName = undef;

    if(  @_ >= 1) {
	$printerName = shift;
    }
    else {
	die " Error !!! printers::getPrinterPPDfilename - printer name required  \n";
    }


    my $printer = $printers->{$printerName} ;

    return $PPDfolder."/".$printer->{ppd};
}

#________________________________________________________________________________
#
#  package:     Support::printers
#
#  subroutine:  getPrinterPAFfilename
#   
#  description: returns the PAF filename assoc. to the printer 
#
#  parameters:  the printer Name 
#
#  return:      the PAF filename
#________________________________________________________________________________
#
sub getPrinterPAFfilename {

    my $this = shift;
    my $PAFfolder = $this->{PAFfolder};
    my $PPDfolder = $this->{PPDfolder};
    my $printers  = $this->{printers};

    my $printerName = undef;

    if(  @_ >= 1) {
	$printerName = shift;
    }
    else {
	die " Error !!! printers::getPrinterPAFfilename - printer name required  \n";
    }


    my $printer = $printers->{$printerName} ;

    return $printer->{paf};
}

#________________________________________________________________________________
#
#  package:     Support::printers
#
#  subroutine:  getPrinterPAFfullname
#   
#  description: returns the PAF filename assoc. to the printer 
#
#  parameters:  the printer Name 
#
#  return:      the PAF filename
#________________________________________________________________________________
#
sub getPrinterPAFfullname {

    my $this = shift;
    my $PAFfolder = $this->{PAFfolder};
    my $PPDfolder = $this->{PPDfolder};
    my $printers  = $this->{printers};

    my $printerName = undef;

    if(  @_ >= 1) {
	$printerName = shift;
    }
    else {
	die " Error !!! printers::getPrinterPAFfilename - printer name required  \n";
    }


    my $printer = $printers->{$printerName} ;

    return $PAFfolder."/".$printer->{paf};
}


1;






